@@ -1,21 +1,20 @@
-# Copyright (c) 2019 Ultimaker B.V.
+# Copyright (c) 2021 Ultimaker B.V.
 # Uranium is released under the terms of the LGPLv3 or higher.
 
-import collections
 import gc
-import os
-import pickle #For serializing/deserializing Python classes to binary files
-import re #For finding containers with asterisks in the constraints and for detecting backup files.
+import re  # For finding containers with asterisks in the constraints and for detecting backup files.
 import time
+import sqlite3 as db
 from typing import Any, cast, Dict, List, Optional, Set, Type, TYPE_CHECKING
-
+import os
 import UM.Dictionary
 import UM.FlameProfiler
 from UM.LockFile import LockFile
 from UM.Logger import Logger
 from UM.MimeTypeDatabase import MimeType, MimeTypeDatabase
-from UM.PluginRegistry import PluginRegistry #To register the container type plug-ins and container provider plug-ins.
+from UM.PluginRegistry import PluginRegistry  # To register the container type plug-ins and container provider plug-ins.
 from UM.Resources import Resources
+from UM.Settings.EmptyInstanceContainer import EmptyInstanceContainer
 from UM.Settings.ContainerFormatError import ContainerFormatError
 from UM.Settings.ContainerProvider import ContainerProvider
 from UM.Settings.constant_instance_containers import empty_container
@@ -25,16 +24,20 @@ from UM.Settings.DefinitionContainer import DefinitionContainer
 from UM.Settings.InstanceContainer import InstanceContainer
 from UM.Settings.Interfaces import ContainerInterface, ContainerRegistryInterface, DefinitionContainerInterface
 from UM.Signal import Signal, signalemitter
+from .DatabaseContainerMetadataController import DatabaseMetadataContainerController
 
 if TYPE_CHECKING:
     from UM.PluginObject import PluginObject
     from UM.Qt.QtApplication import QtApplication
 
 
+metadata_type = Dict[str, Any]
+
+
 @signalemitter
 class ContainerRegistry(ContainerRegistryInterface):
     """Central class to manage all setting providers.
-    
+
     This class aggregates all data from all container providers. If only the
     metadata is used, it requests the metadata lazily from the providers. If
     more than that is needed, the entire container is requested from the
@@ -56,7 +59,7 @@ class ContainerRegistry(ContainerRegistryInterface):
         self._providers = []  # type: List[ContainerProvider]
         PluginRegistry.addType("container_provider", self.addProvider)
 
-        self.metadata = {}  # type: Dict[str, Dict[str, Any]]
+        self.metadata = {}  # type: Dict[str, metadata_type]
         self._containers = {}  # type: Dict[str, ContainerInterface]
         self._wrong_container_ids = set() # type: Set[str]  # Set of already known wrong containers that must be skipped
         self.source_provider = {}  # type: Dict[str, Optional[ContainerProvider]]  # Where each container comes from.
@@ -66,9 +69,19 @@ class ContainerRegistry(ContainerRegistryInterface):
         self.source_provider["empty"] = None
         self._resource_types = {"definition": Resources.DefinitionContainers}  # type: Dict[str, int]
 
-        #Since queries are based on metadata, we need to make sure to clear the cache when a container's metadata changes.
+        # Since queries are based on metadata, we need to make sure to clear the cache when a container's metadata
+        # changes.
         self.containerMetaDataChanged.connect(self._clearQueryCache)
 
+        # We use a database to store the metadata so that we don't have to extract them from the files every time
+        # the application starts. Reading the data from a lot of files is especially slow on Windows; about 30x as slow.
+        self._db_connection: Optional[db.Connection] = None
+
+        # Since each container that we can store in the database has different metadata (and thus needs different logic
+        # to extract it from the database again), we use database controllers to do that. These are set by type; Each
+        # type of container needs to have their own controller.
+        self._database_handlers: Dict[str, DatabaseMetadataContainerController] = {}
+
         self._explicit_read_only_container_ids = set()  # type: Set[str]
 
     containerAdded = Signal()
@@ -104,7 +117,7 @@ class ContainerRegistry(ContainerRegistryInterface):
 
     def findDefinitionContainers(self, **kwargs: Any) -> List[DefinitionContainerInterface]:
         """Find all DefinitionContainer objects matching certain criteria.
-        
+
         :param dict kwargs: A dictionary of keyword arguments containing
         keys and values that need to match the metadata of the
         DefinitionContainer. An asterisk in the values can be used to denote a
@@ -115,7 +128,7 @@ class ContainerRegistry(ContainerRegistryInterface):
 
     def findDefinitionContainersMetadata(self, **kwargs: Any) -> List[Dict[str, Any]]:
         """Get the metadata of all definition containers matching certain criteria.
-        
+
         :param kwargs: A dictionary of keyword arguments containing keys and
         values that need to match the metadata. An asterisk in the values can be
         used to denote a wildcard.
@@ -127,7 +140,7 @@ class ContainerRegistry(ContainerRegistryInterface):
 
     def findInstanceContainers(self, **kwargs: Any) -> List[InstanceContainer]:
         """Find all InstanceContainer objects matching certain criteria.
-        
+
         :param kwargs: A dictionary of keyword arguments containing
         keys and values that need to match the metadata of the
         InstanceContainer. An asterisk in the values can be used to denote a
@@ -136,9 +149,9 @@ class ContainerRegistry(ContainerRegistryInterface):
 
         return cast(List[InstanceContainer], self.findContainers(container_type = InstanceContainer, **kwargs))
 
-    def findInstanceContainersMetadata(self, **kwargs: Any) -> List[Dict[str, Any]]:
+    def findInstanceContainersMetadata(self, **kwargs: Any) -> List[metadata_type]:
         """Find the metadata of all instance containers matching certain criteria.
-        
+
         :param kwargs: A dictionary of keyword arguments containing keys and
         values that need to match the metadata. An asterisk in the values can be
         used to denote a wildcard.
@@ -150,7 +163,7 @@ class ContainerRegistry(ContainerRegistryInterface):
 
     def findContainerStacks(self, **kwargs: Any) -> List[ContainerStack]:
         """Find all ContainerStack objects matching certain criteria.
-        
+
         :param kwargs: A dictionary of keyword arguments containing
         keys and values that need to match the metadata of the ContainerStack.
         An asterisk in the values can be used to denote a wildcard.
@@ -158,9 +171,9 @@ class ContainerRegistry(ContainerRegistryInterface):
 
         return cast(List[ContainerStack], self.findContainers(container_type = ContainerStack, **kwargs))
 
-    def findContainerStacksMetadata(self, **kwargs: Any) -> List[Dict[str, Any]]:
+    def findContainerStacksMetadata(self, **kwargs: Any) -> List[metadata_type]:
         """Find the metadata of all container stacks matching certain criteria.
-        
+
         :param kwargs: A dictionary of keyword arguments containing keys and
         values that need to match the metadata. An asterisk in the values can be
         used to denote a wildcard.
@@ -173,13 +186,13 @@ class ContainerRegistry(ContainerRegistryInterface):
     @UM.FlameProfiler.profile
     def findContainers(self, *, ignore_case: bool = False, **kwargs: Any) -> List[ContainerInterface]:
         """Find all container objects matching certain criteria.
-        
+
         :param container_type: If provided, return only objects that are
         instances or subclasses of container_type.
         :param kwargs: A dictionary of keyword arguments containing
         keys and values that need to match the metadata of the container. An
         asterisk can be used to denote a wildcard.
-        
+
         :return: A list of containers matching the search criteria, or an empty
         list if nothing was found.
         """
@@ -211,9 +224,9 @@ class ContainerRegistry(ContainerRegistryInterface):
                 result.append(new_container)
         return result
 
-    def findContainersMetadata(self, *, ignore_case: bool = False, **kwargs: Any) -> List[Dict[str, Any]]:
+    def findContainersMetadata(self, *, ignore_case: bool = False, **kwargs: Any) -> List[metadata_type]:
         """Find the metadata of all container objects matching certain criteria.
-        
+
         :param container_type: If provided, return only objects that are
         instances or subclasses of ``container_type``.
         :param kwargs: A dictionary of keyword arguments containing keys and
@@ -254,18 +267,18 @@ class ContainerRegistry(ContainerRegistryInterface):
         query = ContainerQuery.ContainerQuery(self, ignore_case = ignore_case, **kwargs)
         query.execute(candidates = candidates)
 
-        return cast(List[Dict[str, Any]], query.getResult())  # As the execute of the query is done, result won't be none.
+        return cast(List[metadata_type], query.getResult())  # As the execute of the query is done, result won't be none.
 
     def findDirtyContainers(self, *, ignore_case: bool = False, **kwargs: Any) -> List[ContainerInterface]:
         """Specialized find function to find only the modified container objects
         that also match certain criteria.
-        
+
         This is faster than the normal find methods since it won't ever load all
         containers, but only the modified ones. Since containers must be fully
         loaded before they are modified, you are guaranteed that any operations
         on the resulting containers will not trigger additional containers to
         load lazily.
-        
+
         :param kwargs: A dictionary of keyword arguments containing
         keys and values that need to match the metadata of the container. An
         asterisk can be used to denote a wildcard.
@@ -305,7 +318,7 @@ class ContainerRegistry(ContainerRegistryInterface):
 
     def isReadOnly(self, container_id: str) -> bool:
         """Returns whether a profile is read-only or not.
-        
+
         Whether it is read-only depends on the source where the container is
         obtained from.
         :return: True if the container is read-only, or False if it can be
@@ -329,7 +342,7 @@ class ContainerRegistry(ContainerRegistryInterface):
 
     def isLoaded(self, container_id: str) -> bool:
         """Returns whether a container is completely loaded or not.
-        
+
         If only its metadata is known, it is not yet completely loaded.
         :return: True if all data about this container is known, False if only
         metadata is known or the container is completely unknown.
@@ -337,35 +350,188 @@ class ContainerRegistry(ContainerRegistryInterface):
 
         return container_id in self._containers
 
+    def _createDatabaseFile(self, db_path: str) -> db.Connection:
+        connection = db.Connection(db_path)
+        cursor = connection.cursor()
+        cursor.executescript("""
+            CREATE TABLE containers(
+                id text,
+                name text,
+                last_modified integer,
+                container_type text
+            );
+            CREATE UNIQUE INDEX idx_containers_id on containers (id);
+        """)
+
+        for handler in self._database_handlers.values():
+            handler.setupTable(cursor)
+
+        return connection
+
+    def _getDatabaseConnection(self) -> db.Connection:
+        if self._db_connection is not None:
+            return self._db_connection
+        db_path = os.path.join(Resources.getCacheStoragePath(), "containers.db")
+        if not os.path.exists(db_path):
+            self._db_connection = self._createDatabaseFile(db_path)
+            return self._db_connection
+        self._db_connection = db.Connection(db_path)
+        return self._db_connection
+
+    def _getProfileType(self, container_id: str, db_cursor: db.Cursor) -> Optional[str]:
+        db_cursor.execute("select id, container_type from containers where id = ?", (container_id, ))
+        row = db_cursor.fetchone()
+        if row:
+            return row[1]
+        return None
+
+    def _recreateCorruptDataBase(self, cursor: Optional[db.Cursor]) -> None:
+        """Closes the Database, removes the file from cache and recreate all metadata from scratch"""
+        if not cursor:
+            self.loadAllMetadata()
+            return
+        try:
+            cursor.execute("rollback")  # Cancel any ongoing transaction.
+        except:
+            # Could be that the cursor is already closed
+            pass
+        cursor.close()
+
+        self._db_connection = None
+        db_path = os.path.join(Resources.getCacheStoragePath(), "containers.db")
+        try:
+            os.remove(db_path)
+        except EnvironmentError:  # Was already deleted by rollback.
+            pass
+
+        self.loadAllMetadata()
+
+    def _getProfileModificationTime(self, container_id: str, db_cursor: db.Cursor) -> Optional[float]:
+        db_cursor.execute("select id, last_modified from containers where id = ?", (container_id, ))
+        row = db_cursor.fetchone()
+
+        if row:
+            return row[1]
+        return None
+
+    def _addMetadataToDatabase(self, metadata: metadata_type) -> None:
+        container_type = metadata["type"]
+        if container_type in self._database_handlers:
+            try:
+                self._database_handlers[container_type].insert(metadata)
+            except db.DatabaseError as e:
+                Logger.warning(f"Removing corrupt database and recreating database. {e}")
+                self._recreateCorruptDataBase(self._database_handlers[container_type].cursor)
+
+    def _updateMetadataInDatabase(self, metadata: metadata_type) -> None:
+        container_type = metadata["type"]
+        if container_type in self._database_handlers:
+            try:
+                self._database_handlers[container_type].update(metadata)
+            except db.DatabaseError as e:
+                Logger.warning(f"Removing corrupt database and recreating database. {e}")
+                self._recreateCorruptDataBase(self._database_handlers[container_type].cursor)
+
+    def _getMetadataFromDatabase(self, container_id: str, container_type: str) -> metadata_type:
+        if container_type in self._database_handlers:
+            return self._database_handlers[container_type].getMetadata(container_id)
+        return {}
+
     def loadAllMetadata(self) -> None:
         """Load the metadata of all available definition containers, instance
         containers and container stacks.
         """
 
+        cursor = self._getDatabaseConnection().cursor()
+        for handlers in self._database_handlers.values():
+            handlers.cursor = cursor
+
         self._clearQueryCache()
         gc.disable()
         resource_start_time = time.time()
+
+        # Since it could well be that we have to make a *lot* of changes to the database, we want to do that in
+        # a single transaction to speed it up.
+        cursor.execute("begin")
+        all_container_ids = set()
         for provider in self._providers:  # Automatically sorted by the priority queue.
-            for container_id in list(provider.getAllIds()):  # Make copy of all IDs since it might change during iteration.
-                if container_id not in self.metadata:
-                    self._application.processEvents()  # Update the user interface because loading takes a while. Specifically the loading screen.
+            # Make copy of all IDs since it might change during iteration.
+            provider_container_ids = set(provider.getAllIds())
+            # Keep a list of all the ID's that we know off
+            all_container_ids.update(provider_container_ids)
+            for container_id in provider_container_ids:
+                try:
+                    db_last_modified_time = self._getProfileModificationTime(container_id, cursor)
+                except db.DatabaseError as e:
+                    Logger.warning(f"Removing corrupt database and recreating database. {e}")
+                    self._recreateCorruptDataBase(cursor)
+                    cursor = self._getDatabaseConnection().cursor()  # After recreating the database, all the cursors have changed.
+                    cursor.execute("begin")
+                    db_last_modified_time = self._getProfileModificationTime(container_id, cursor)
+                if db_last_modified_time is None:
+                    # Item is not yet in the database. Add it now!
                     metadata = provider.loadMetadata(container_id)
                     if not self._isMetadataValid(metadata):
-                        Logger.log("w", "Invalid metadata for container {container_id}: {metadata}".format(container_id = container_id, metadata = metadata))
-                        if container_id in self.metadata:
-                            del self.metadata[container_id]
+                        Logger.log("w", f"Invalid metadata for container {container_id}: {metadata}")
                         continue
+                    modified_time = provider.getLastModifiedTime(container_id)
+                    if metadata.get("type") in self._database_handlers:
+                        # Only add it to the database if we have an actual handler.
+                        cursor.execute(
+                            "INSERT INTO containers (id, name, last_modified, container_type) VALUES (?, ?, ?, ?)",
+                            (container_id, metadata["name"], modified_time, metadata["type"]))
+                        self._addMetadataToDatabase(metadata)
+
                     self.metadata[container_id] = metadata
                     self.source_provider[container_id] = provider
+
+                else:
+                    # Metadata already exists in database.
+                    modified_time = provider.getLastModifiedTime(container_id)
+                    if modified_time > db_last_modified_time:
+                        # Metadata is outdated, so load from file and update the database
+                        metadata = provider.loadMetadata(container_id)
+                        cursor.execute("UPDATE containers SET name = ?, last_modified = ?, container_type = ? WHERE id = ?", (metadata["name"], modified_time, metadata["type"], metadata["id"]))
+                        self._updateMetadataInDatabase(metadata)
+                        self.metadata[container_id] = metadata
+                        self.source_provider[container_id] = provider
+                        continue
+
+                    # Since we know that the container exists, we also know that it will never be None.
+                    container_type = cast(str, self._getProfileType(container_id, cursor))
+
+                    # No need to do any file reading, we can just get it from the database.
+                    self.metadata[container_id] = self._getMetadataFromDatabase(container_id, container_type)
+                    self.source_provider[container_id] = provider
+
+        cursor.execute("commit")
+
+        # Find all ID's that we currently have in the database
+        cursor.execute("SELECT id from containers")
+        all_ids_in_database = {container_id[0] for container_id in cursor.fetchall()}
+        ids_to_remove = all_ids_in_database - all_container_ids
+
+        # Purge ID's that don't have a matching file
+        for container_id in ids_to_remove:
+            cursor.execute("DELETE FROM containers WHERE id = ?", (container_id,))
+            self._removeContainerFromDatabase(container_id)
+
+        if ids_to_remove:  # We only can (and need to) commit again if we removed containers
+            cursor.execute("commit")
+
         Logger.log("d", "Loading metadata into container registry took %s seconds", time.time() - resource_start_time)
         gc.enable()
         ContainerRegistry.allMetadataLoaded.emit()
 
+    def _removeContainerFromDatabase(self, container_id: str) -> None:
+        for database_handler in self._database_handlers.values():
+            database_handler.delete(container_id)
+
     @UM.FlameProfiler.profile
     def load(self) -> None:
         """Load all available definition containers, instance containers and
         container stacks.
-        
+
         :note This method does not clear the internal list of containers. This means that any containers
         that were already added when the first call to this method happened will not be re-added.
         """
@@ -393,10 +559,10 @@ class ContainerRegistry(ContainerRegistryInterface):
         Logger.log("d", "Loading data into container registry took %s seconds", time.time() - resource_start_time)
 
     @UM.FlameProfiler.profile
-    def addContainer(self, container: ContainerInterface) -> None:
+    def addContainer(self, container: ContainerInterface) -> bool:
         container_id = container.getId()
         if container_id in self._containers:
-            return
+            return True  # Container was already there, consider that a success
 
         if hasattr(container, "metaDataChanged"):
             container.metaDataChanged.connect(self._onContainerMetaDataChanged)
@@ -404,7 +570,7 @@ class ContainerRegistry(ContainerRegistryInterface):
         self.metadata[container_id] = container.getMetaData()
         self._containers[container_id] = container
         if container_id not in self.source_provider:
-            self.source_provider[container_id] = None #Added during runtime.
+            self.source_provider[container_id] = None  # Added during runtime.
         self._clearQueryCacheByContainer(container)
 
         # containerAdded is a custom signal and can trigger direct calls to its subscribers. This should be avoided
@@ -413,6 +579,7 @@ class ContainerRegistry(ContainerRegistryInterface):
         # We avoid the direct calls here to make sure that the subscribers do not need to take into account any max
         # recursion problem.
         self._application.callLater(self.containerAdded.emit, container)
+        return True
 
     @UM.FlameProfiler.profile
     def removeContainer(self, container_id: str) -> None:
@@ -441,12 +608,11 @@ class ContainerRegistry(ContainerRegistryInterface):
         if container_id in self.metadata:
             if container is None:
                 # We're in a bit of a weird state now. We want to notify the rest of the code that the container
-                # has been deleted, but due to lazy loading, it hasnt even been loaded yet. The issues is that in order
-                # to notify the rest of the code, we need to actually *have* the container. So we need to load it
-                # in order to remove it...
-                provider = self.source_provider.get(container_id)
-                if provider:
-                    container = provider.loadContainer(container_id)
+                # has been deleted, but due to lazy loading, it hasn't been loaded yet. The issues is that in order
+                # to notify the rest of the code, we need to actually *have* the container. So an empty instance 
+                # container is created, which is emitted with the containerRemoved signal and contains the metadata
+                container = EmptyInstanceContainer(container_id)
+                container.metaData = self.metadata[container_id]
             del self.metadata[container_id]
         if container_id in self.source_provider:
             if self.source_provider[container_id] is not None:
@@ -500,10 +666,10 @@ class ContainerRegistry(ContainerRegistryInterface):
     @UM.FlameProfiler.profile
     def uniqueName(self, original: str) -> str:
         """Creates a new unique name for a container that doesn't exist yet.
-        
+
         It tries if the original name you provide exists, and if it doesn't
         it'll add a "        1" or "        2" after the name to make it unique.
-        
+
         :param original: The original name that may not be unique.
         :return: A unique name that looks a lot like the original but may have
         a number behind it to make it unique.
@@ -532,7 +698,7 @@ class ContainerRegistry(ContainerRegistryInterface):
     @classmethod
     def addContainerType(cls, container: "PluginObject") -> None:
         """Add a container type that will be used to serialize/deserialize containers.
-        
+
         :param container: An instance of the container type to add.
         """
 
@@ -556,9 +722,9 @@ class ContainerRegistry(ContainerRegistryInterface):
     @classmethod
     def getMimeTypeForContainer(cls, container_type: type) -> Optional[MimeType]:
         """Retrieve the mime type corresponding to a certain container type
-        
+
         :param container_type: The type of container to get the mime type for.
-        
+
         :return: A MimeType object that matches the mime type of the container or None if not found.
         """
 
@@ -571,11 +737,11 @@ class ContainerRegistry(ContainerRegistryInterface):
         return None
 
     @classmethod
-    def getContainerForMimeType(cls, mime_type):
+    def getContainerForMimeType(cls, mime_type: MimeType) -> Optional[Type[ContainerInterface]]:
         """Get the container type corresponding to a certain mime type.
-        
+
         :param mime_type: The mime type to get the container type for.
-        
+
         :return: A class object of a container type that corresponds to the specified mime type or None if not found.
         """
 
@@ -584,7 +750,7 @@ class ContainerRegistry(ContainerRegistryInterface):
     @classmethod
     def getContainerTypes(cls):
         """Get all the registered container types
-        
+
         :return: A dictionary view object that provides access to the container types.
         The key is the plugin ID, the value the container type.
         """
@@ -600,6 +766,7 @@ class ContainerRegistry(ContainerRegistryInterface):
             return
 
         provider.saveContainer(container) #type: ignore
+        container.setDirty(False)
         self.source_provider[container.getId()] = provider
 
     def saveDirtyContainers(self) -> None:
@@ -613,53 +780,6 @@ class ContainerRegistry(ContainerRegistryInterface):
             for stack in self.findContainerStacks():
                 self.saveContainer(stack)
 
-    # Load a binary cached version of a DefinitionContainer
-    def _loadCachedDefinition(self, definition_id: str, path: str) -> None:
-        try:
-            cache_path = Resources.getPath(Resources.Cache, "definitions", self._application.getVersion(), definition_id)
-
-            cache_mtime = os.path.getmtime(cache_path)
-            definition_mtime = os.path.getmtime(path)
-
-            if definition_mtime > cache_mtime:
-                # The definition is newer than the cached version, so ignore the cached version.
-                Logger.log("d", "Definition file %s is newer than cache, ignoring cached version", path)
-                return None
-
-            with open(cache_path, "rb") as f:
-                definition = pickle.load(f)
-
-            for file_path in definition.getInheritedFiles():
-                if os.path.getmtime(file_path) > cache_mtime:
-                    return None
-
-            return definition
-        except FileNotFoundError:
-            return None
-        except Exception as e:
-            # We could not load a cached version for some reason. Ignore it.
-            Logger.logException("d", "Could not load cached definition for {path}: {err}".format(path = path, err = str(e)))
-            return None
-
-    # Store a cached version of a DefinitionContainer
-    def _saveCachedDefinition(self, definition: DefinitionContainer):
-        cache_path = Resources.getStoragePath(Resources.Cache, "definitions", self._application.getVersion(), definition.id)
-
-        # Ensure the cache path exists
-        os.makedirs(os.path.dirname(cache_path), exist_ok = True)
-
-        try:
-            with open(cache_path, "wb") as f:
-                pickle.dump(definition, f, pickle.HIGHEST_PROTOCOL)
-        except RecursionError:
-            #Sometimes a recursion error in pickling occurs here.
-            #The cause is unknown. It must be some circular reference in the definition instances or definition containers.
-            #Instead of saving a partial cache and raising an exception, simply fail to save the cache.
-            #See CURA-4024.
-            Logger.log("w", "The definition cache for definition {definition_id} failed to pickle.".format(definition_id = definition.getId()))
-            if os.path.exists(cache_path):
-                os.remove(cache_path) #The pickling might be half-complete, which causes EOFError in Pickle when you load it later.
-
     # Clear the internal query cache
     def _clearQueryCache(self, *args: Any, **kwargs: Any) -> None:
         ContainerQuery.ContainerQuery.cache.clear()
@@ -684,7 +804,7 @@ class ContainerRegistry(ContainerRegistryInterface):
 
     def _onContainerMetaDataChanged(self, *args: ContainerInterface, **kwargs: Any) -> None:
         """Called when any container's metadata changed.
-        
+
         This function passes it on to the containerMetaDataChanged signal. Sadly
         that doesn't work automatically between pyqtSignal and UM.Signal.
         """
@@ -694,9 +814,9 @@ class ContainerRegistry(ContainerRegistryInterface):
         self.metadata[container.getId()] = container.getMetaData()  # refresh the metadata
         self.containerMetaDataChanged.emit(*args, **kwargs)
 
-    def _isMetadataValid(self, metadata: Optional[Dict[str, Any]]) -> bool:
+    def _isMetadataValid(self, metadata: Optional[metadata_type]) -> bool:
         """Validate a metadata object.
-        
+
         If the metadata is invalid, the container is not allowed to be in the
         registry.
         :param metadata: A metadata object.
